@@ -43,11 +43,11 @@ FILE_LICENCE ( GPL2_OR_LATER );
 #include "config/serial.h"
 
 /** The "SYSLINUX" version string */
-static char __data16_array ( syslinux_version, [] ) = "gPXE " VERSION;
+static char __data16_array ( syslinux_version, [] ) = "\r\ngPXE " VERSION;
 #define syslinux_version __use_data16 ( syslinux_version )
 
 /** The "SYSLINUX" copyright string */
-static char __data16_array ( syslinux_copyright, [] ) = "http://etherboot.org";
+static char __data16_array ( syslinux_copyright, [] ) = " http://etherboot.org";
 #define syslinux_copyright __use_data16 ( syslinux_copyright )
 
 static char __data16_array ( syslinux_configuration_file, [] ) = "";
@@ -318,6 +318,20 @@ static __asmcall void int21 ( struct i386_all_regs *ix86 ) {
 }
 
 
+/**
+ * Dispatch PXE API call weakly
+ *
+ * @v ix86		Registers for PXE call
+ * @ret present		Zero if the PXE stack is present, nonzero if not
+ *
+ * A successful return only indicates that the PXE stack was available
+ * for dispatching the call; it says nothing about the success of
+ * whatever the call asked for.
+ */
+__weak int pxe_api_call_weak ( struct i386_all_regs *ix86 __unused ) {
+	return -1;
+}
+
 /**
  * SYSLINUX API
  */
@@ -445,8 +459,10 @@ static __asmcall void int22 ( struct i386_all_regs *ix86 ) {
 		break;
 
 	case 0x0009: /* Call PXE Stack */
-		pxe_api_call ( ix86 );
-		ix86->flags &= ~CF;
+		if ( pxe_api_call_weak ( ix86 ) != 0 )
+			ix86->flags |= CF;
+		else
+			ix86->flags &= ~CF;
 		break;
 
 	case 0x000A: /* Get Derivative-Specific Information */
@@ -667,6 +683,7 @@ void hook_comboot_interrupts ( ) {
 		              "pushw %%cs\n\t"
 		              "call prot_call\n\t"
 		              "addw $4, %%sp\n\t"
+			      "call patch_cf\n\t"
 		              "iret\n\t" )
 		          : : "i" ( int20 ) );
 
@@ -679,6 +696,7 @@ void hook_comboot_interrupts ( ) {
 		              "pushw %%cs\n\t"
 		              "call prot_call\n\t"
 		              "addw $4, %%sp\n\t"
+			      "call patch_cf\n\t"
 		              "iret\n\t" )
 		          : : "i" ( int21 ) );
 
@@ -691,6 +709,7 @@ void hook_comboot_interrupts ( ) {
 		              "pushw %%cs\n\t"
 		              "call prot_call\n\t"
 		              "addw $4, %%sp\n\t"
+			      "call patch_cf\n\t"
 		              "iret\n\t" )
 		          : : "i" ( int22) );
 
