@@ -16,6 +16,8 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
+FILE_LICENCE ( GPL2_OR_LATER );
+
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
@@ -29,6 +31,9 @@
 #include <gpxe/efi/Protocol/DriverBinding.h>
 #include <gpxe/efi/Protocol/PciIo.h>
 #include <gpxe/efi/Protocol/SimpleNetwork.h>
+#include <gpxe/efi/Protocol/ComponentName2.h>
+#include <gpxe/efi/Protocol/NetworkInterfaceIdentifier.h>
+#include <config/general.h>
 
 /** @file
  *
@@ -40,6 +45,8 @@
 struct efi_snp_device {
 	/** The underlying gPXE network device */
 	struct net_device *netdev;
+	/** EFI device handle */
+	EFI_HANDLE handle;
 	/** The SNP structure itself */
 	EFI_SIMPLE_NETWORK_PROTOCOL snp;
 	/** The SNP "mode" (parameters) */
@@ -58,6 +65,16 @@ struct efi_snp_device {
 	unsigned int rx_count_interrupts;
 	/** Outstanding RX packet count (via WaitForPacket event) */
 	unsigned int rx_count_events;
+	/** The network interface identifier */
+	EFI_NETWORK_INTERFACE_IDENTIFIER_PROTOCOL nii;
+	/** Device name */
+	wchar_t name[ sizeof ( ( ( struct net_device * ) NULL )->name ) ];
+	/** The device path
+	 *
+	 * This field is variable in size and must appear at the end
+	 * of the structure.
+	 */
+	EFI_DEVICE_PATH_PROTOCOL path;
 };
 
 /** EFI simple network protocol GUID */
@@ -68,6 +85,28 @@ static EFI_GUID efi_simple_network_protocol_guid
 static EFI_GUID efi_driver_binding_protocol_guid
 	= EFI_DRIVER_BINDING_PROTOCOL_GUID;
 
+/** EFI component name protocol GUID */
+static EFI_GUID efi_component_name2_protocol_guid
+	= EFI_COMPONENT_NAME2_PROTOCOL_GUID;
+
+/** EFI device path protocol GUID */
+static EFI_GUID efi_device_path_protocol_guid
+	= EFI_DEVICE_PATH_PROTOCOL_GUID;
+
+/** EFI network interface identifier GUID */
+static EFI_GUID efi_nii_protocol_guid
+	= EFI_NETWORK_INTERFACE_IDENTIFIER_PROTOCOL_GUID;
+
+/** EFI network interface identifier GUID (extra special version) */
+static EFI_GUID efi_nii31_protocol_guid = {
+	/* At some point, it seems that someone decided to change the
+	 * GUID.  Current EFI builds ignore the older GUID, older EFI
+	 * builds ignore the newer GUID, so we have to expose both.
+	 */
+	0x1ACED566, 0x76ED, 0x4218,
+	{ 0xBC, 0x81, 0x76, 0x7F, 0x1F, 0x97, 0x7A, 0x89 }
+};
+
 /** EFI PCI I/O protocol GUID */
 static EFI_GUID efi_pci_io_protocol_guid
 	= EFI_PCI_IO_PROTOCOL_GUID;
@@ -80,20 +119,20 @@ static EFI_GUID efi_pci_io_protocol_guid
 static void efi_snp_set_mode ( struct efi_snp_device *snpdev ) {
 	struct net_device *netdev = snpdev->netdev;
 	EFI_SIMPLE_NETWORK_MODE *mode = &snpdev->mode;
-	unsigned int ll_addr_len = netdev->ll_protocol->ll_addr_len;
+	struct ll_protocol *ll_protocol = netdev->ll_protocol;
+	unsigned int ll_addr_len = ll_protocol->ll_addr_len;
 
 	mode->HwAddressSize = ll_addr_len;
-	mode->MediaHeaderSize = netdev->ll_protocol->ll_header_len;
+	mode->MediaHeaderSize = ll_protocol->ll_header_len;
 	mode->MaxPacketSize = netdev->max_pkt_len;
 	mode->ReceiveFilterMask = ( EFI_SIMPLE_NETWORK_RECEIVE_UNICAST |
 				    EFI_SIMPLE_NETWORK_RECEIVE_MULTICAST |
 				    EFI_SIMPLE_NETWORK_RECEIVE_BROADCAST );
 	assert ( ll_addr_len <= sizeof ( mode->CurrentAddress ) );
 	memcpy ( &mode->CurrentAddress, netdev->ll_addr, ll_addr_len );
-	memcpy ( &mode->BroadcastAddress, netdev->ll_protocol->ll_broadcast,
-		 ll_addr_len );
-	memcpy ( &mode->PermanentAddress, netdev->ll_addr, ll_addr_len );
-	mode->IfType = ntohs ( netdev->ll_protocol->ll_proto );
+	memcpy ( &mode->BroadcastAddress, netdev->ll_broadcast, ll_addr_len );
+	ll_protocol->init_addr ( netdev->hw_addr, &mode->PermanentAddress );
+	mode->IfType = ntohs ( ll_protocol->ll_proto );
 	mode->MacAddressChangeable = TRUE;
 	mode->MediaPresentSupported = TRUE;
 	mode->MediaPresent = ( netdev_link_ok ( netdev ) ? TRUE : FALSE );
@@ -289,7 +328,7 @@ efi_snp_station_address ( EFI_SIMPLE_NETWORK_PROTOCOL *snp, BOOLEAN reset,
 	memcpy ( snpdev->netdev->ll_addr, new, ll_protocol->ll_addr_len );
 
 	/* MAC address changes take effect only on netdev_open() */
-	if ( snpdev->netdev->state & NETDEV_OPEN ) {
+	if ( netdev_is_open ( snpdev->netdev ) ) {
 		DBGC ( snpdev, "SNPDEV %p MAC address changed while net "
 		       "devive open\n", snpdev );
 	}
@@ -555,7 +594,8 @@ efi_snp_transmit ( EFI_SIMPLE_NETWORK_PROTOCOL *snp,
 	/* Create link-layer header, if specified */
 	if ( ll_header_len ) {
 		iob_pull ( iobuf, ll_header_len );
-		if ( ( rc = ll_protocol->push ( iobuf, ll_dest, ll_src,
+		if ( ( rc = ll_protocol->push ( snpdev->netdev,
+						iobuf, ll_dest, ll_src,
 						htons ( *net_proto ) )) != 0 ){
 			DBGC ( snpdev, "SNPDEV %p TX could not construct "
 			       "header: %s\n", snpdev, strerror ( rc ) );
@@ -565,10 +605,9 @@ efi_snp_transmit ( EFI_SIMPLE_NETWORK_PROTOCOL *snp,
 	}
 
 	/* Transmit packet */
-	if ( ( rc = netdev_tx ( snpdev->netdev, iobuf ) ) != 0 ) {
+	if ( ( rc = netdev_tx ( snpdev->netdev, iob_disown ( iobuf ) ) ) != 0){
 		DBGC ( snpdev, "SNPDEV %p TX could not transmit: %s\n",
 		       snpdev, strerror ( rc ) );
-		iobuf = NULL;
 		efirc = RC_TO_EFIRC ( rc );
 		goto err_tx;
 	}
@@ -634,8 +673,8 @@ efi_snp_receive ( EFI_SIMPLE_NETWORK_PROTOCOL *snp,
 	*len = iob_len ( iobuf );
 
 	/* Attempt to decode link-layer header */
-	if ( ( rc = ll_protocol->pull ( iobuf, &iob_ll_dest, &iob_ll_src,
-					&iob_net_proto ) ) != 0 ) {
+	if ( ( rc = ll_protocol->pull ( snpdev->netdev, iobuf, &iob_ll_dest,
+					&iob_ll_src, &iob_net_proto ) ) != 0 ){
 		DBGC ( snpdev, "SNPDEV %p could not parse header: %s\n",
 		       snpdev, strerror ( rc ) );
 		efirc = RC_TO_EFIRC ( rc );
@@ -674,7 +713,7 @@ static VOID EFIAPI efi_snp_wait_for_packet ( EFI_EVENT event,
 	DBGCP ( snpdev, "SNPDEV %p WAIT_FOR_PACKET\n", snpdev );
 
 	/* Do nothing unless the net device is open */
-	if ( ! ( snpdev->netdev->state & NETDEV_OPEN ) )
+	if ( ! netdev_is_open ( snpdev->netdev ) )
 		return;
 
 	/* Poll the network device */
@@ -784,7 +823,7 @@ efi_snp_snpdev ( EFI_DRIVER_BINDING_PROTOCOL *driver, EFI_HANDLE device ) {
 		EFI_SIMPLE_NETWORK_PROTOCOL *snp;
 		void *interface;
 	} u;
-	struct efi_snp_device *snpdev;
+	struct efi_snp_device *snpdev = NULL;
 	EFI_STATUS efirc;
 
 	if ( ( efirc = bs->OpenProtocol ( device,
@@ -795,12 +834,16 @@ efi_snp_snpdev ( EFI_DRIVER_BINDING_PROTOCOL *driver, EFI_HANDLE device ) {
 					  EFI_OPEN_PROTOCOL_GET_PROTOCOL))!=0){
 		DBGC ( driver, "SNPDRV %p device %p could not locate SNP: "
 		       "%s\n", driver, device, efi_strerror ( efirc ) );
-		return NULL;
+		goto err_no_snp;
 	}
 
 	snpdev =  container_of ( u.snp, struct efi_snp_device, snp );
 	DBGCP ( driver, "SNPDRV %p device %p is SNPDEV %p\n",
 		driver, device, snpdev );
+
+	bs->CloseProtocol ( device, &efi_simple_network_protocol_guid,
+			    driver->DriverBindingHandle, device );
+ err_no_snp:
 	return snpdev;
 }
 
@@ -838,15 +881,41 @@ efi_snp_driver_start ( EFI_DRIVER_BINDING_PROTOCOL *driver,
 		       EFI_HANDLE device,
 		       EFI_DEVICE_PATH_PROTOCOL *child ) {
 	EFI_BOOT_SERVICES *bs = efi_systab->BootServices;
+	EFI_DEVICE_PATH_PROTOCOL *path;
+	EFI_DEVICE_PATH_PROTOCOL *subpath;
+	MAC_ADDR_DEVICE_PATH *macpath;
 	struct efi_snp_device *snpdev;
 	struct net_device *netdev;
+	size_t subpath_len;
+	size_t path_prefix_len = 0;
+	unsigned int i;
 	EFI_STATUS efirc;
 
 	DBGCP ( driver, "SNPDRV %p DRIVER_START %p (%p)\n",
 		driver, device, child );
 
+	/* Determine device path prefix length */
+	if ( ( efirc = bs->OpenProtocol ( device,
+					  &efi_device_path_protocol_guid,
+					  ( void * ) &path,
+					  driver->DriverBindingHandle,
+					  device,
+					  EFI_OPEN_PROTOCOL_BY_DRIVER )) !=0 ){
+		DBGCP ( driver, "SNPDRV %p device %p has no device path\n",
+			driver, device );
+		goto err_no_device_path;
+	}
+	subpath = path;
+	while ( subpath->Type != END_DEVICE_PATH_TYPE ) {
+		subpath_len = ( ( subpath->Length[1] << 8 ) |
+				subpath->Length[0] );
+		path_prefix_len += subpath_len;
+		subpath = ( ( ( void * ) subpath ) + subpath_len );
+	}
+
 	/* Allocate the SNP device */
-	snpdev = zalloc ( sizeof ( *snpdev ) );
+	snpdev = zalloc ( sizeof ( *snpdev ) + path_prefix_len +
+			  sizeof ( *macpath ) );
 	if ( ! snpdev ) {
 		efirc = EFI_OUT_OF_RESOURCES;
 		goto err_alloc_snp;
@@ -886,22 +955,60 @@ efi_snp_driver_start ( EFI_DRIVER_BINDING_PROTOCOL *driver,
 	snpdev->mode.State = EfiSimpleNetworkStopped;
 	efi_snp_set_mode ( snpdev );
 
+	/* Populate the NII structure */
+	snpdev->nii.Revision =
+		EFI_NETWORK_INTERFACE_IDENTIFIER_PROTOCOL_REVISION;
+	strncpy ( snpdev->nii.StringId, "gPXE",
+		  sizeof ( snpdev->nii.StringId ) );
+
+	/* Populate the device name */
+	for ( i = 0 ; i < sizeof ( netdev->name ) ; i++ ) {
+		/* Damn Unicode names */
+		assert ( i < ( sizeof ( snpdev->name ) /
+			       sizeof ( snpdev->name[0] ) ) );
+		snpdev->name[i] = netdev->name[i];
+	}
+
+	/* Populate the device path */
+	memcpy ( &snpdev->path, path, path_prefix_len );
+	macpath = ( ( ( void * ) &snpdev->path ) + path_prefix_len );
+	subpath = ( ( void * ) ( macpath + 1 ) );
+	memset ( macpath, 0, sizeof ( *macpath ) );
+	macpath->Header.Type = MESSAGING_DEVICE_PATH;
+	macpath->Header.SubType = MSG_MAC_ADDR_DP;
+	macpath->Header.Length[0] = sizeof ( *macpath );
+	memcpy ( &macpath->MacAddress, netdev->ll_addr,
+		 sizeof ( macpath->MacAddress ) );
+	macpath->IfType = ntohs ( netdev->ll_protocol->ll_proto );
+	memset ( subpath, 0, sizeof ( *subpath ) );
+	subpath->Type = END_DEVICE_PATH_TYPE;
+	subpath->SubType = END_ENTIRE_DEVICE_PATH_SUBTYPE;
+	subpath->Length[0] = sizeof ( *subpath );
+
 	/* Install the SNP */
-	if ( ( efirc = bs->InstallProtocolInterface ( &device,
-				&efi_simple_network_protocol_guid,
-				EFI_NATIVE_INTERFACE, &snpdev->snp ) ) != 0 ) {
-		DBGC ( snpdev, "SNPDEV %p could not install protocol: %s\n",
-		       snpdev, efi_strerror ( efirc ) );
+	if ( ( efirc = bs->InstallMultipleProtocolInterfaces (
+			&snpdev->handle,
+			&efi_simple_network_protocol_guid, &snpdev->snp,
+			&efi_device_path_protocol_guid, &snpdev->path,
+			&efi_nii_protocol_guid, &snpdev->nii,
+			&efi_nii31_protocol_guid, &snpdev->nii,
+			NULL ) ) != 0 ) {
+		DBGC ( snpdev, "SNPDEV %p could not install protocols: "
+		       "%s\n", snpdev, efi_strerror ( efirc ) );
 		goto err_install_protocol_interface;
 	}
 
-	DBGC ( snpdev, "SNPDEV %p installed for %s on device %p\n",
-	       snpdev, netdev->name, device );
+	DBGC ( snpdev, "SNPDEV %p installed for %s as device %p\n",
+	       snpdev, netdev->name, snpdev->handle );
 	return 0;
 
-	bs->UninstallProtocolInterface ( device,
-					 &efi_simple_network_protocol_guid,
-					 &snpdev->snp );
+	bs->UninstallMultipleProtocolInterfaces (
+			snpdev->handle,
+			&efi_simple_network_protocol_guid, &snpdev->snp,
+			&efi_device_path_protocol_guid, &snpdev->path,
+			&efi_nii_protocol_guid, &snpdev->nii,
+			&efi_nii31_protocol_guid, &snpdev->nii,
+			NULL );
  err_install_protocol_interface:
 	bs->CloseEvent ( snpdev->snp.WaitForPacket );
  err_create_event:
@@ -910,6 +1017,9 @@ efi_snp_driver_start ( EFI_DRIVER_BINDING_PROTOCOL *driver,
  err_no_netdev:
 	free ( snpdev );
  err_alloc_snp:
+	bs->CloseProtocol ( device, &efi_device_path_protocol_guid,
+			    driver->DriverBindingHandle, device );
+ err_no_device_path:
 	return efirc;
 }
 
@@ -942,12 +1052,18 @@ efi_snp_driver_stop ( EFI_DRIVER_BINDING_PROTOCOL *driver,
 	}
 
 	/* Uninstall the SNP */
-	bs->UninstallProtocolInterface ( device,
-					 &efi_simple_network_protocol_guid,
-					 &snpdev->snp );
+	bs->UninstallMultipleProtocolInterfaces (
+			snpdev->handle,
+			&efi_simple_network_protocol_guid, &snpdev->snp,
+			&efi_device_path_protocol_guid, &snpdev->path,
+			&efi_nii_protocol_guid, &snpdev->nii,
+			&efi_nii31_protocol_guid, &snpdev->nii,
+			NULL );
 	bs->CloseEvent ( snpdev->snp.WaitForPacket );
 	netdev_put ( snpdev->netdev );
 	free ( snpdev );
+	bs->CloseProtocol ( device, &efi_device_path_protocol_guid,
+			    driver->DriverBindingHandle, device );
 	return 0;
 }
 
@@ -961,6 +1077,50 @@ static EFI_DRIVER_BINDING_PROTOCOL efi_snp_binding = {
 	NULL
 };
 
+/**
+ * Look up driver name
+ *
+ * @v wtf		Component name protocol
+ * @v language		Language to use
+ * @v driver_name	Driver name to fill in
+ * @ret efirc		EFI status code
+ */
+static EFI_STATUS EFIAPI
+efi_snp_get_driver_name ( EFI_COMPONENT_NAME2_PROTOCOL *wtf __unused,
+			  CHAR8 *language __unused, CHAR16 **driver_name ) {
+
+	*driver_name = L"" PRODUCT_SHORT_NAME " Driver";
+	return 0;
+}
+
+/**
+ * Look up controller name
+ *
+ * @v wtf		Component name protocol
+ * @v device		Device
+ * @v child		Child device, or NULL
+ * @v language		Language to use
+ * @v driver_name	Device name to fill in
+ * @ret efirc		EFI status code
+ */
+static EFI_STATUS EFIAPI
+efi_snp_get_controller_name ( EFI_COMPONENT_NAME2_PROTOCOL *wtf __unused,
+			      EFI_HANDLE device __unused,
+			      EFI_HANDLE child __unused,
+			      CHAR8 *language __unused,
+			      CHAR16 **controller_name __unused ) {
+
+	/* Just let EFI use the default Device Path Name */
+	return EFI_UNSUPPORTED;
+}
+
+/** EFI SNP component name protocol */
+static EFI_COMPONENT_NAME2_PROTOCOL efi_snp_name = {
+	efi_snp_get_driver_name,
+	efi_snp_get_controller_name,
+	"en"
+};
+
 /**
  * Install EFI SNP driver
  *
@@ -972,12 +1132,12 @@ int efi_snp_install ( void ) {
 	EFI_STATUS efirc;
 
 	driver->ImageHandle = efi_image_handle;
-	if ( ( efirc = bs->InstallProtocolInterface (
-					&driver->DriverBindingHandle,
-					&efi_driver_binding_protocol_guid,
-					EFI_NATIVE_INTERFACE,
-					driver ) ) != 0 ) {
-		DBGC ( driver, "SNPDRV %p could not install driver binding: "
+	if ( ( efirc = bs->InstallMultipleProtocolInterfaces (
+			&driver->DriverBindingHandle,
+			&efi_driver_binding_protocol_guid, driver,
+			&efi_component_name2_protocol_guid, &efi_snp_name,
+			NULL ) ) != 0 ) {
+		DBGC ( driver, "SNPDRV %p could not install protocols: "
 		       "%s\n", driver, efi_strerror ( efirc ) );
 		return EFIRC_TO_RC ( efirc );
 	}
