@@ -16,6 +16,8 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
+FILE_LICENCE ( GPL2_OR_LATER );
+
 #include <stddef.h>
 #include <string.h>
 #include <stdlib.h>
@@ -31,8 +33,10 @@
 #include <gpxe/process.h>
 #include <gpxe/uaccess.h>
 #include <gpxe/tcpip.h>
-#include <gpxe/dhcp.h>
+#include <gpxe/settings.h>
 #include <gpxe/features.h>
+#include <gpxe/base16.h>
+#include <gpxe/base64.h>
 #include <gpxe/iscsi.h>
 
 /** @file
@@ -43,17 +47,36 @@
 
 FEATURE ( FEATURE_PROTOCOL, "iSCSI", DHCP_EB_FEATURE_ISCSI, 1 );
 
+/* Disambiguate the various error causes */
+#define EACCES_INCORRECT_TARGET_USERNAME ( EACCES | EUNIQ_01 )
+#define EACCES_INCORRECT_TARGET_PASSWORD ( EACCES | EUNIQ_02 )
+#define ENOTSUP_INITIATOR_STATUS ( ENOTSUP | EUNIQ_01 )
+#define ENOTSUP_OPCODE ( ENOTSUP | EUNIQ_02 )
+#define ENOTSUP_DISCOVERY ( ENOTSUP | EUNIQ_03 )
+#define EPERM_INITIATOR_AUTHENTICATION ( EPERM | EUNIQ_01 )
+#define EPERM_INITIATOR_AUTHORISATION ( EPERM | EUNIQ_02 )
+#define EPROTO_INVALID_CHAP_ALGORITHM ( EPROTO | EUNIQ_01 )
+#define EPROTO_INVALID_CHAP_IDENTIFIER ( EPROTO | EUNIQ_02 )
+#define EPROTO_INVALID_LARGE_BINARY ( EPROTO | EUNIQ_03 )
+#define EPROTO_INVALID_CHAP_RESPONSE ( EPROTO | EUNIQ_04 )
+
 /** iSCSI initiator name (explicitly specified) */
 static char *iscsi_explicit_initiator_iqn;
 
 /** Default iSCSI initiator name (constructed from hostname) */
 static char *iscsi_default_initiator_iqn;
 
-/** iSCSI username */
-static char *iscsi_username;
+/** iSCSI initiator username */
+static char *iscsi_initiator_username;
 
-/** iSCSI password */
-static char *iscsi_password;
+/** iSCSI initiator password */
+static char *iscsi_initiator_password;
+
+/** iSCSI target username */
+static char *iscsi_target_username;
+
+/** iSCSI target password */
+static char *iscsi_target_password;
 
 static void iscsi_start_tx ( struct iscsi_session *iscsi );
 static void iscsi_start_login ( struct iscsi_session *iscsi );
@@ -81,8 +104,10 @@ static void iscsi_free ( struct refcnt *refcnt ) {
 
 	free ( iscsi->target_address );
 	free ( iscsi->target_iqn );
-	free ( iscsi->username );
-	free ( iscsi->password );
+	free ( iscsi->initiator_username );
+	free ( iscsi->initiator_password );
+	free ( iscsi->target_username );
+	free ( iscsi->target_password );
 	chap_finish ( &iscsi->chap );
 	iscsi_rx_buffered_data_done ( iscsi );
 	free ( iscsi );
@@ -117,6 +142,8 @@ static int iscsi_open_connection ( struct iscsi_session *iscsi ) {
 	/* Enter security negotiation phase */
 	iscsi->status = ( ISCSI_STATUS_SECURITY_NEGOTIATION_PHASE |
 			  ISCSI_STATUS_STRINGS_SECURITY );
+	if ( iscsi->target_username )
+		iscsi->status |= ISCSI_STATUS_AUTH_REVERSE_REQUIRED;
 
 	/* Assign fresh initiator task tag */
 	iscsi->itt++;
@@ -170,9 +197,10 @@ static void iscsi_close_connection ( struct iscsi_session *iscsi, int rc ) {
 static void iscsi_scsi_done ( struct iscsi_session *iscsi, int rc ) {
 
 	assert ( iscsi->tx_state == ISCSI_TX_IDLE );
+	assert ( iscsi->command != NULL );
 
+	iscsi->command->rc = rc;
 	iscsi->command = NULL;
-	iscsi->rc = rc;
 }
 
 /****************************************************************************
@@ -213,11 +241,12 @@ static void iscsi_start_command ( struct iscsi_session *iscsi ) {
 	command->cmdsn = htonl ( iscsi->cmdsn );
 	command->expstatsn = htonl ( iscsi->statsn + 1 );
 	memcpy ( &command->cdb, &iscsi->command->cdb, sizeof ( command->cdb ));
-	DBGC ( iscsi, "iSCSI %p start " SCSI_CDB_FORMAT " %s %#zx\n",
-	       iscsi, SCSI_CDB_DATA ( command->cdb ),
-	       ( iscsi->command->data_in ? "in" : "out" ),
-	       ( iscsi->command->data_in ?
-		 iscsi->command->data_in_len : iscsi->command->data_out_len ));
+	DBGC2 ( iscsi, "iSCSI %p start " SCSI_CDB_FORMAT " %s %#zx\n",
+		iscsi, SCSI_CDB_DATA ( command->cdb ),
+		( iscsi->command->data_in ? "in" : "out" ),
+		( iscsi->command->data_in ?
+		  iscsi->command->data_in_len :
+		  iscsi->command->data_out_len ) );
 }
 
 /**
@@ -420,9 +449,9 @@ static int iscsi_tx_data_out ( struct iscsi_session *iscsi ) {
  *
  *     HeaderDigest=None
  *     DataDigest=None
- *     MaxConnections is irrelevant; we make only one connection anyway
+ *     MaxConnections is irrelevant; we make only one connection anyway [4]
  *     InitialR2T=Yes [1]
- *     ImmediateData is irrelevant; we never send immediate data
+ *     ImmediateData is irrelevant; we never send immediate data [4]
  *     MaxRecvDataSegmentLength=8192 (default; we don't care) [3]
  *     MaxBurstLength=262144 (default; we don't care) [3]
  *     FirstBurstLength=262144 (default; we don't care)
@@ -445,43 +474,67 @@ static int iscsi_tx_data_out ( struct iscsi_session *iscsi ) {
  * these parameters, but some targets (notably OpenSolaris)
  * incorrectly assume a default value of zero, so we explicitly
  * specify the default values.
+ *
+ * [4] We are quite happy to use the RFC-defined default values for
+ * these parameters, but some targets (notably a QNAP TS-639Pro) fail
+ * unless they are supplied, so we explicitly specify the default
+ * values.
  */
 static int iscsi_build_login_request_strings ( struct iscsi_session *iscsi,
 					       void *data, size_t len ) {
 	unsigned int used = 0;
-	unsigned int i;
+	const char *auth_method;
 
 	if ( iscsi->status & ISCSI_STATUS_STRINGS_SECURITY ) {
+		/* Default to allowing no authentication */
+		auth_method = "None";
+		/* If we have a credential to supply, permit CHAP */
+		if ( iscsi->initiator_username )
+			auth_method = "CHAP,None";
+		/* If we have a credential to check, force CHAP */
+		if ( iscsi->target_username )
+			auth_method = "CHAP";
 		used += ssnprintf ( data + used, len - used,
 				    "InitiatorName=%s%c"
 				    "TargetName=%s%c"
 				    "SessionType=Normal%c"
-				    "AuthMethod=CHAP,None%c",
+				    "AuthMethod=%s%c",
 				    iscsi_initiator_iqn(), 0,
-				    iscsi->target_iqn, 0, 0, 0 );
+				    iscsi->target_iqn, 0, 0,
+				    auth_method, 0 );
 	}
 
 	if ( iscsi->status & ISCSI_STATUS_STRINGS_CHAP_ALGORITHM ) {
 		used += ssnprintf ( data + used, len - used, "CHAP_A=5%c", 0 );
 	}
 	
-	if ( ( iscsi->status & ISCSI_STATUS_STRINGS_CHAP_RESPONSE ) &&
-	     iscsi->username ) {
+	if ( ( iscsi->status & ISCSI_STATUS_STRINGS_CHAP_RESPONSE ) ) {
+		char buf[ base16_encoded_len ( iscsi->chap.response_len ) + 1 ];
+		assert ( iscsi->initiator_username != NULL );
+		base16_encode ( iscsi->chap.response, iscsi->chap.response_len,
+				buf );
 		used += ssnprintf ( data + used, len - used,
-				    "CHAP_N=%s%cCHAP_R=0x",
-				    iscsi->username, 0 );
-		for ( i = 0 ; i < iscsi->chap.response_len ; i++ ) {
-			used += ssnprintf ( data + used, len - used, "%02x",
-					    iscsi->chap.response[i] );
-		}
-		used += ssnprintf ( data + used, len - used, "%c", 0 );
+				    "CHAP_N=%s%cCHAP_R=0x%s%c",
+				    iscsi->initiator_username, 0, buf, 0 );
+	}
+
+	if ( ( iscsi->status & ISCSI_STATUS_STRINGS_CHAP_CHALLENGE ) ) {
+		size_t challenge_len = ( sizeof ( iscsi->chap_challenge ) - 1 );
+		char buf[ base16_encoded_len ( challenge_len ) + 1 ];
+		base16_encode ( ( iscsi->chap_challenge + 1 ), challenge_len,
+				buf );
+		used += ssnprintf ( data + used, len - used,
+				    "CHAP_I=%d%cCHAP_C=0x%s%c",
+				    iscsi->chap_challenge[0], 0, buf, 0 );
 	}
 
 	if ( iscsi->status & ISCSI_STATUS_STRINGS_OPERATIONAL ) {
 		used += ssnprintf ( data + used, len - used,
 				    "HeaderDigest=None%c"
 				    "DataDigest=None%c"
+				    "MaxConnections=1%c"
 				    "InitialR2T=Yes%c"
+				    "ImmediateData=No%c"
 				    "MaxRecvDataSegmentLength=8192%c"
 				    "MaxBurstLength=262144%c"
 				    "DefaultTime2Wait=0%c"
@@ -490,7 +543,7 @@ static int iscsi_build_login_request_strings ( struct iscsi_session *iscsi,
 				    "DataPDUInOrder=Yes%c"
 				    "DataSequenceInOrder=Yes%c"
 				    "ErrorRecoveryLevel=0%c",
-				    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 );
+				    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 );
 	}
 
 	return used;
@@ -561,6 +614,41 @@ static int iscsi_tx_login_request ( struct iscsi_session *iscsi ) {
 	return xfer_deliver_iob ( &iscsi->socket, iobuf );
 }
 
+/**
+ * Calculate maximum length of decoded large binary value
+ *
+ * @v encoded		Encoded large binary value
+ * @v max_raw_len	Maximum length of raw data
+ */
+static inline size_t
+iscsi_large_binary_decoded_max_len ( const char *encoded ) {
+	return ( strlen ( encoded ) ); /* Decoding never expands data */
+}
+
+/**
+ * Decode large binary value
+ *
+ * @v encoded		Encoded large binary value
+ * @v raw		Raw data
+ * @ret len		Length of raw data, or negative error
+ */
+static int iscsi_large_binary_decode ( const char *encoded, uint8_t *raw ) {
+
+	if ( encoded[0] != '0' )
+		return -EPROTO_INVALID_LARGE_BINARY;
+
+	switch ( encoded[1] ) {
+	case 'x' :
+	case 'X' :
+		return base16_decode ( ( encoded + 2 ), raw );
+	case 'b' :
+	case 'B' :
+		return base64_decode ( ( encoded + 2 ), raw );
+	default:
+		return -EPROTO_INVALID_LARGE_BINARY;
+	}
+}
+
 /**
  * Handle iSCSI TargetAddress text value
  *
@@ -605,8 +693,10 @@ static int iscsi_handle_authmethod_value ( struct iscsi_session *iscsi,
 	if ( strcmp ( value, "CHAP" ) == 0 ) {
 		DBGC ( iscsi, "iSCSI %p initiating CHAP authentication\n",
 		       iscsi );
-		iscsi->status |= ISCSI_STATUS_STRINGS_CHAP_ALGORITHM;
+		iscsi->status |= ( ISCSI_STATUS_STRINGS_CHAP_ALGORITHM |
+				   ISCSI_STATUS_AUTH_FORWARD_REQUIRED );
 	}
+
 	return 0;
 }
 
@@ -619,7 +709,6 @@ static int iscsi_handle_authmethod_value ( struct iscsi_session *iscsi,
  */
 static int iscsi_handle_chap_a_value ( struct iscsi_session *iscsi,
 				       const char *value ) {
-	int rc;
 
 	/* We only ever offer "5" (i.e. MD5) as an algorithm, so if
 	 * the server responds with anything else it is a protocol
@@ -628,14 +717,7 @@ static int iscsi_handle_chap_a_value ( struct iscsi_session *iscsi,
 	if ( strcmp ( value, "5" ) != 0 ) {
 		DBGC ( iscsi, "iSCSI %p got invalid CHAP algorithm \"%s\"\n",
 		       iscsi, value );
-		return -EPROTO;
-	}
-
-	/* Prepare for CHAP with MD5 */
-	if ( ( rc = chap_init ( &iscsi->chap, &md5_algorithm ) ) != 0 ) {
-		DBGC ( iscsi, "iSCSI %p could not initialise CHAP: %s\n",
-		       iscsi, strerror ( rc ) );
-		return rc;
+		return -EPROTO_INVALID_CHAP_ALGORITHM;
 	}
 
 	return 0;
@@ -652,22 +734,31 @@ static int iscsi_handle_chap_i_value ( struct iscsi_session *iscsi,
 				       const char *value ) {
 	unsigned int identifier;
 	char *endp;
+	int rc;
 
 	/* The CHAP identifier is an integer value */
 	identifier = strtoul ( value, &endp, 0 );
 	if ( *endp != '\0' ) {
 		DBGC ( iscsi, "iSCSI %p saw invalid CHAP identifier \"%s\"\n",
 		       iscsi, value );
-		return -EPROTO;
+		return -EPROTO_INVALID_CHAP_IDENTIFIER;
+	}
+
+	/* Prepare for CHAP with MD5 */
+	chap_finish ( &iscsi->chap );
+	if ( ( rc = chap_init ( &iscsi->chap, &md5_algorithm ) ) != 0 ) {
+		DBGC ( iscsi, "iSCSI %p could not initialise CHAP: %s\n",
+		       iscsi, strerror ( rc ) );
+		return rc;
 	}
 
 	/* Identifier and secret are the first two components of the
 	 * challenge.
 	 */
 	chap_set_identifier ( &iscsi->chap, identifier );
-	if ( iscsi->password ) {
-		chap_update ( &iscsi->chap, iscsi->password,
-			      strlen ( iscsi->password ) );
+	if ( iscsi->initiator_password ) {
+		chap_update ( &iscsi->chap, iscsi->initiator_password,
+			      strlen ( iscsi->initiator_password ) );
 	}
 
 	return 0;
@@ -682,35 +773,120 @@ static int iscsi_handle_chap_i_value ( struct iscsi_session *iscsi,
  */
 static int iscsi_handle_chap_c_value ( struct iscsi_session *iscsi,
 				       const char *value ) {
-	char buf[3];
-	char *endp;
-	uint8_t byte;
+	uint8_t buf[ iscsi_large_binary_decoded_max_len ( value ) ];
+	unsigned int i;
+	size_t len;
+	int rc;
 
-	/* Check and strip leading "0x" */
-	if ( ( value[0] != '0' ) || ( value[1] != 'x' ) ) {
-		DBGC ( iscsi, "iSCSI %p saw invalid CHAP challenge \"%s\"\n",
-		       iscsi, value );
-	}
-	value += 2;
-
-	/* Process challenge an octet at a time */
-	for ( ; ( value[0] && value[1] ) ; value += 2 ) {
-		memcpy ( buf, value, 2 );
-		buf[2] = 0;
-		byte = strtoul ( buf, &endp, 16 );
-		if ( *endp != '\0' ) {
-			DBGC ( iscsi, "iSCSI %p saw invalid CHAP challenge "
-			       "byte \"%s\"\n", iscsi, buf );
-			return -EPROTO;
-		}
-		chap_update ( &iscsi->chap, &byte, sizeof ( byte ) );
+	/* Process challenge */
+	rc = iscsi_large_binary_decode ( value, buf );
+	if ( rc < 0 ) {
+		DBGC ( iscsi, "iSCSI %p invalid CHAP challenge \"%s\": %s\n",
+		       iscsi, value, strerror ( rc ) );
+		return rc;
 	}
+	len = rc;
+	chap_update ( &iscsi->chap, buf, len );
 
 	/* Build CHAP response */
 	DBGC ( iscsi, "iSCSI %p sending CHAP response\n", iscsi );
 	chap_respond ( &iscsi->chap );
 	iscsi->status |= ISCSI_STATUS_STRINGS_CHAP_RESPONSE;
 
+	/* Send CHAP challenge, if applicable */
+	if ( iscsi->target_username ) {
+		iscsi->status |= ISCSI_STATUS_STRINGS_CHAP_CHALLENGE;
+		/* Generate CHAP challenge data */
+		for ( i = 0 ; i < sizeof ( iscsi->chap_challenge ) ; i++ ) {
+			iscsi->chap_challenge[i] = random();
+		}
+	}
+
+	return 0;
+}
+
+/**
+ * Handle iSCSI CHAP_N text value
+ *
+ * @v iscsi		iSCSI session
+ * @v value		CHAP_N value
+ * @ret rc		Return status code
+ */
+static int iscsi_handle_chap_n_value ( struct iscsi_session *iscsi,
+				       const char *value ) {
+
+	/* The target username isn't actually involved at any point in
+	 * the authentication process; it merely serves to identify
+	 * which password the target is using to generate the CHAP
+	 * response.  We unnecessarily verify that the username is as
+	 * expected, in order to provide mildly helpful diagnostics if
+	 * the target is supplying the wrong username/password
+	 * combination.
+	 */
+	if ( iscsi->target_username &&
+	     ( strcmp ( iscsi->target_username, value ) != 0 ) ) {
+		DBGC ( iscsi, "iSCSI %p target username \"%s\" incorrect "
+		       "(wanted \"%s\")\n",
+		       iscsi, value, iscsi->target_username );
+		return -EACCES_INCORRECT_TARGET_USERNAME;
+	}
+
+	return 0;
+}
+
+/**
+ * Handle iSCSI CHAP_R text value
+ *
+ * @v iscsi		iSCSI session
+ * @v value		CHAP_R value
+ * @ret rc		Return status code
+ */
+static int iscsi_handle_chap_r_value ( struct iscsi_session *iscsi,
+				       const char *value ) {
+	uint8_t buf[ iscsi_large_binary_decoded_max_len ( value ) ];
+	size_t len;
+	int rc;
+
+	/* Generate CHAP response for verification */
+	chap_finish ( &iscsi->chap );
+	if ( ( rc = chap_init ( &iscsi->chap, &md5_algorithm ) ) != 0 ) {
+		DBGC ( iscsi, "iSCSI %p could not initialise CHAP: %s\n",
+		       iscsi, strerror ( rc ) );
+		return rc;
+	}
+	chap_set_identifier ( &iscsi->chap, iscsi->chap_challenge[0] );
+	if ( iscsi->target_password ) {
+		chap_update ( &iscsi->chap, iscsi->target_password,
+			      strlen ( iscsi->target_password ) );
+	}
+	chap_update ( &iscsi->chap, &iscsi->chap_challenge[1],
+		      ( sizeof ( iscsi->chap_challenge ) - 1 ) );
+	chap_respond ( &iscsi->chap );
+
+	/* Process response */
+	rc = iscsi_large_binary_decode ( value, buf );
+	if ( rc < 0 ) {
+		DBGC ( iscsi, "iSCSI %p invalid CHAP response \"%s\": %s\n",
+		       iscsi, value, strerror ( rc ) );
+		return rc;
+	}
+	len = rc;
+
+	/* Check CHAP response */
+	if ( len != iscsi->chap.response_len ) {
+		DBGC ( iscsi, "iSCSI %p invalid CHAP response length\n",
+		       iscsi );
+		return -EPROTO_INVALID_CHAP_RESPONSE;
+	}
+	if ( memcmp ( buf, iscsi->chap.response, len ) != 0 ) {
+		DBGC ( iscsi, "iSCSI %p incorrect CHAP response \"%s\"\n",
+		       iscsi, value );
+		return -EACCES_INCORRECT_TARGET_PASSWORD;
+	}
+
+	/* Mark session as authenticated */
+	iscsi->status |= ISCSI_STATUS_AUTH_REVERSE_OK;
+
 	return 0;
 }
 
@@ -738,6 +914,8 @@ static struct iscsi_string_type iscsi_string_types[] = {
 	{ "CHAP_A=", iscsi_handle_chap_a_value },
 	{ "CHAP_I=", iscsi_handle_chap_i_value },
 	{ "CHAP_C=", iscsi_handle_chap_c_value },
+	{ "CHAP_N=", iscsi_handle_chap_n_value },
+	{ "CHAP_R=", iscsi_handle_chap_r_value },
 	{ NULL, NULL }
 };
 
@@ -829,6 +1007,35 @@ static int iscsi_rx_buffered_data ( struct iscsi_session *iscsi,
 	return 0;
 }
 
+/**
+ * Convert iSCSI response status to return status code
+ *
+ * @v status_class	iSCSI status class
+ * @v status_detail	iSCSI status detail
+ * @ret rc		Return status code
+ */
+static int iscsi_status_to_rc ( unsigned int status_class,
+				unsigned int status_detail ) {
+	switch ( status_class ) {
+	case ISCSI_STATUS_INITIATOR_ERROR :
+		switch ( status_detail ) {
+		case ISCSI_STATUS_INITIATOR_ERROR_AUTHENTICATION :
+			return -EPERM_INITIATOR_AUTHENTICATION;
+		case ISCSI_STATUS_INITIATOR_ERROR_AUTHORISATION :
+			return -EPERM_INITIATOR_AUTHORISATION;
+		case ISCSI_STATUS_INITIATOR_ERROR_NOT_FOUND :
+		case ISCSI_STATUS_INITIATOR_ERROR_REMOVED :
+			return -ENODEV;
+		default :
+			return -ENOTSUP_INITIATOR_STATUS;
+		}
+	case ISCSI_STATUS_TARGET_ERROR :
+		return -EIO;
+	default :
+		return -EINVAL;
+	}
+}
+
 /**
  * Receive data segment of an iSCSI login response PDU
  *
@@ -876,20 +1083,24 @@ static int iscsi_rx_login_response ( struct iscsi_session *iscsi,
 	if ( response->status_class != 0 ) {
 		DBGC ( iscsi, "iSCSI login failure: class %02x detail %02x\n",
 		       response->status_class, response->status_detail );
-		iscsi->instant_rc = -EPERM;
-		return -EPERM;
+		rc = iscsi_status_to_rc ( response->status_class,
+					  response->status_detail );
+		iscsi->instant_rc = rc;
+		return rc;
 	}
 
 	/* Handle login transitions */
 	if ( response->flags & ISCSI_LOGIN_FLAG_TRANSITION ) {
+		iscsi->status &= ~( ISCSI_STATUS_PHASE_MASK |
+				    ISCSI_STATUS_STRINGS_MASK );
 		switch ( response->flags & ISCSI_LOGIN_NSG_MASK ) {
 		case ISCSI_LOGIN_NSG_OPERATIONAL_NEGOTIATION:
-			iscsi->status =
+			iscsi->status |=
 				( ISCSI_STATUS_OPERATIONAL_NEGOTIATION_PHASE |
 				  ISCSI_STATUS_STRINGS_OPERATIONAL );
 			break;
 		case ISCSI_LOGIN_NSG_FULL_FEATURE_PHASE:
-			iscsi->status = ISCSI_STATUS_FULL_FEATURE_PHASE;
+			iscsi->status |= ISCSI_STATUS_FULL_FEATURE_PHASE;
 			break;
 		default:
 			DBGC ( iscsi, "iSCSI %p got invalid response flags "
@@ -907,6 +1118,14 @@ static int iscsi_rx_login_response ( struct iscsi_session *iscsi,
 		return 0;
 	}
 
+	/* Check that target authentication was successful (if required) */
+	if ( ( iscsi->status & ISCSI_STATUS_AUTH_REVERSE_REQUIRED ) &&
+	     ! ( iscsi->status & ISCSI_STATUS_AUTH_REVERSE_OK ) ) {
+		DBGC ( iscsi, "iSCSI %p nefarious target tried to bypass "
+		       "authentication\n", iscsi );
+		return -EPROTO;
+	}
+
 	/* Reset retry count */
 	iscsi->retry_count = 0;
 
@@ -1116,9 +1335,9 @@ static int iscsi_rx_bhs ( struct iscsi_session *iscsi, const void *data,
 			  size_t len, size_t remaining __unused ) {
 	memcpy ( &iscsi->rx_bhs.bytes[iscsi->rx_offset], data, len );
 	if ( ( iscsi->rx_offset + len ) >= sizeof ( iscsi->rx_bhs ) ) {
-		DBGC ( iscsi, "iSCSI %p received PDU opcode %#x len %#lx\n",
-		       iscsi, iscsi->rx_bhs.common.opcode,
-		       ISCSI_DATA_LEN ( iscsi->rx_bhs.common.lengths ) );
+		DBGC2 ( iscsi, "iSCSI %p received PDU opcode %#x len %#x\n",
+			iscsi, iscsi->rx_bhs.common.opcode,
+			ISCSI_DATA_LEN ( iscsi->rx_bhs.common.lengths ) );
 	}
 	return 0;
 }
@@ -1176,7 +1395,7 @@ static int iscsi_rx_data ( struct iscsi_session *iscsi, const void *data,
 			return 0;
 		DBGC ( iscsi, "iSCSI %p unknown opcode %02x\n", iscsi,
 		       response->opcode );
-		return -EOPNOTSUPP;
+		return -ENOTSUP_OPCODE;
 	}
 }
 
@@ -1326,7 +1545,7 @@ static int iscsi_vredirect ( struct xfer_interface *socket, int type,
 		va_end ( tmp );
 	}
 
-	return xfer_vopen ( socket, type, args );
+	return xfer_vreopen ( socket, type, args );
 }
 			     
 
@@ -1360,37 +1579,24 @@ static int iscsi_command ( struct scsi_device *scsi,
 		container_of ( scsi->backend, struct iscsi_session, refcnt );
 	int rc;
 
+	/* Abort immediately if we have a recorded permanent failure */
+	if ( iscsi->instant_rc )
+		return iscsi->instant_rc;
+
 	/* Record SCSI command */
 	iscsi->command = command;
 
-	/* Abort immediately if we have a recorded permanent failure */
-	if ( iscsi->instant_rc ) {
-		rc = iscsi->instant_rc;
-		goto done;
-	}
-
 	/* Issue command or open connection as appropriate */
 	if ( iscsi->status ) {
 		iscsi_start_command ( iscsi );
 	} else {
-		if ( ( rc = iscsi_open_connection ( iscsi ) ) != 0 )
-			goto done;
+		if ( ( rc = iscsi_open_connection ( iscsi ) ) != 0 ) {
+			iscsi->command = NULL;
+			return rc;
+		}
 	}
 
-	/* Wait for command to complete */
-	iscsi->rc = -EINPROGRESS;
-	while ( iscsi->rc == -EINPROGRESS )
-		step();
-	rc = iscsi->rc;
-
- done:
-	iscsi->command = NULL;
-	return rc;
-}
-
-static int iscsi_detached_command ( struct scsi_device *scsi __unused,
-				    struct scsi_command *command __unused ) {
-	return -ENODEV;
+	return 0;
 }
 
 /**
@@ -1405,7 +1611,7 @@ void iscsi_detach ( struct scsi_device *scsi ) {
 	xfer_nullify ( &iscsi->socket );
 	iscsi_close_connection ( iscsi, 0 );
 	process_del ( &iscsi->process );
-	scsi->command = iscsi_detached_command;
+	scsi->command = scsi_detached_command;
 	ref_put ( scsi->backend );
 	scsi->backend = NULL;
 }
@@ -1427,39 +1633,6 @@ enum iscsi_root_path_component {
 	NUM_RP_COMPONENTS
 };
 
-/**
- * Parse iSCSI LUN
- *
- * @v iscsi		iSCSI session
- * @v lun_string	LUN string representation (as per RFC4173)
- * @ret rc		Return status code
- */
-static int iscsi_parse_lun ( struct iscsi_session *iscsi,
-			     const char *lun_string ) {
-	char *p = ( char * ) lun_string;
-	union {
-		uint64_t u64;
-		uint16_t u16[4];
-	} lun;
-	int i;
-
-	/* Empty LUN; assume LUN 0 */
-	if ( ! *lun_string )
-		return 0;
-
-	for ( i = 0 ; i < 4 ; i++ ) {
-		lun.u16[i] = strtoul ( p, &p, 16 );
-		if ( *p != '-' )
-			return -EINVAL;
-		p++;
-	}
-	if ( *p )
-		return -EINVAL;
-
-	iscsi->lun = lun.u64;
-	return 0;
-}
-
 /**
  * Parse iSCSI root path
  *
@@ -1498,7 +1671,7 @@ static int iscsi_parse_root_path ( struct iscsi_session *iscsi,
 	iscsi->target_port = strtoul ( rp_comp[RP_PORT], NULL, 10 );
 	if ( ! iscsi->target_port )
 		iscsi->target_port = ISCSI_PORT;
-	if ( ( rc = iscsi_parse_lun ( iscsi, rp_comp[RP_LUN] ) ) != 0 ) {
+	if ( ( rc = scsi_parse_lun ( rp_comp[RP_LUN], &iscsi->lun ) ) != 0 ) {
 		DBGC ( iscsi, "iSCSI %p invalid LUN \"%s\"\n",
 		       iscsi, rp_comp[RP_LUN] );
 		return rc;
@@ -1514,26 +1687,61 @@ static int iscsi_parse_root_path ( struct iscsi_session *iscsi,
  * Set iSCSI authentication details
  *
  * @v iscsi		iSCSI session
- * @v username		Username, if any
- * @v password		Password, if any
+ * @v initiator_username Initiator username, if any
+ * @v initiator_password Initiator password, if any
+ * @v target_username	Target username, if any
+ * @v target_password	Target password, if any
  * @ret rc		Return status code
  */
 static int iscsi_set_auth ( struct iscsi_session *iscsi,
-			    const char *username, const char *password ) {
+			    const char *initiator_username,
+			    const char *initiator_password,
+			    const char *target_username,
+			    const char *target_password ) {
 
-	if ( username ) {
-		iscsi->username = strdup ( username );
-		if ( ! iscsi->username )
-			return -ENOMEM;
-	}
+	/* Check for initiator or target credentials */
+	if ( initiator_username || initiator_password ||
+	     target_username || target_password ) {
 
-	if ( password ) {
-		iscsi->password = strdup ( password );
-		if ( ! iscsi->password )
+		/* We must have at least an initiator username+password */
+		if ( ! ( initiator_username && initiator_password ) )
+			goto invalid_auth;
+
+		/* Store initiator credentials */
+		iscsi->initiator_username = strdup ( initiator_username );
+		if ( ! iscsi->initiator_username )
 			return -ENOMEM;
+		iscsi->initiator_password = strdup ( initiator_password );
+		if ( ! iscsi->initiator_password )
+			return -ENOMEM;
+
+		/* Check for target credentials */
+		if ( target_username || target_password ) {
+
+			/* We must have target username+password */
+			if ( ! ( target_username && target_password ) )
+				goto invalid_auth;
+
+			/* Store target credentials */
+			iscsi->target_username = strdup ( target_username );
+			if ( ! iscsi->target_username )
+				return -ENOMEM;
+			iscsi->target_password = strdup ( target_password );
+			if ( ! iscsi->target_password )
+				return -ENOMEM;
+		}
 	}
 
 	return 0;
+
+ invalid_auth:
+	DBGC ( iscsi, "iSCSI %p invalid credentials: initiator "
+	       "%sname,%spw, target %sname,%spw\n", iscsi,
+	       ( initiator_username ? "" : "no " ),
+	       ( initiator_password ? "" : "no " ),
+	       ( target_username ? "" : "no " ),
+	       ( target_password ? "" : "no " ) );
+	return -EINVAL;
 }
 
 /**
@@ -1559,15 +1767,18 @@ int iscsi_attach ( struct scsi_device *scsi, const char *root_path ) {
 	if ( ( rc = iscsi_parse_root_path ( iscsi, root_path ) ) != 0 )
 		goto err;
 	/* Set fields not specified by root path */
-	if ( ( rc = iscsi_set_auth ( iscsi, iscsi_username,
-				     iscsi_password ) ) != 0 )
+	if ( ( rc = iscsi_set_auth ( iscsi,
+				     iscsi_initiator_username,
+				     iscsi_initiator_password,
+				     iscsi_target_username,
+				     iscsi_target_password ) ) != 0 )
 		goto err;
 
 	/* Sanity checks */
 	if ( ! iscsi->target_address ) {
 		DBGC ( iscsi, "iSCSI %p does not yet support discovery\n",
 		       iscsi );
-		rc = -ENOTSUP;
+		rc = -ENOTSUP_DISCOVERY;
 		goto err;
 	}
 	if ( ! iscsi->target_iqn ) {
@@ -1580,7 +1791,6 @@ int iscsi_attach ( struct scsi_device *scsi, const char *root_path ) {
 	/* Attach parent interface, mortalise self, and return */
 	scsi->backend = ref_get ( &iscsi->refcnt );
 	scsi->command = iscsi_command;
-	scsi->lun = iscsi->lun;
 	ref_put ( &iscsi->refcnt );
 	return 0;
 	
@@ -1591,78 +1801,143 @@ int iscsi_attach ( struct scsi_device *scsi, const char *root_path ) {
 
 /****************************************************************************
  *
- * DHCP option applicators
+ * Settings
  *
  */
 
+/** iSCSI initiator IQN setting */
+struct setting initiator_iqn_setting __setting = {
+	.name = "initiator-iqn",
+	.description = "iSCSI initiator name",
+	.tag = DHCP_ISCSI_INITIATOR_IQN,
+	.type = &setting_type_string,
+};
+
+/** iSCSI reverse username setting */
+struct setting reverse_username_setting __setting = {
+	.name = "reverse-username",
+	.description = "Reverse user name",
+	.tag = DHCP_EB_REVERSE_USERNAME,
+	.type = &setting_type_string,
+};
+
+/** iSCSI reverse password setting */
+struct setting reverse_password_setting __setting = {
+	.name = "reverse-password",
+	.description = "Reverse password",
+	.tag = DHCP_EB_REVERSE_PASSWORD,
+	.type = &setting_type_string,
+};
+
+/** An iSCSI string setting */
+struct iscsi_string_setting {
+	/** Setting */
+	struct setting *setting;
+	/** String to update */
+	char **string;
+	/** String prefix */
+	const char *prefix;
+};
+
+/** iSCSI string settings */
+static struct iscsi_string_setting iscsi_string_settings[] = {
+	{
+		.setting = &initiator_iqn_setting,
+		.string = &iscsi_explicit_initiator_iqn,
+		.prefix = "",
+	},
+	{
+		.setting = &username_setting,
+		.string = &iscsi_initiator_username,
+		.prefix = "",
+	},
+	{
+		.setting = &password_setting,
+		.string = &iscsi_initiator_password,
+		.prefix = "",
+	},
+	{
+		.setting = &reverse_username_setting,
+		.string = &iscsi_target_username,
+		.prefix = "",
+	},
+	{
+		.setting = &reverse_password_setting,
+		.string = &iscsi_target_password,
+		.prefix = "",
+	},
+	{
+		.setting = &hostname_setting,
+		.string = &iscsi_default_initiator_iqn,
+		.prefix = "iqn.2000-01.org.etherboot:",
+	},
+};
+
 /**
- * Apply DHCP iSCSI option
+ * Apply iSCSI setting
  *
- * @v tag		DHCP option tag
- * @v option		DHCP option
+ * @v setting		iSCSI string setting
  * @ret rc		Return status code
  */
-static int apply_dhcp_iscsi_string ( unsigned int tag,
-				     struct dhcp_option *option ) {
-	char *prefix = "";
+static int apply_iscsi_string_setting ( struct iscsi_string_setting *setting ){
 	size_t prefix_len;
+	int setting_len;
 	size_t len;
-	char **string;
+	int check_len;
 	char *p;
 
-	/* Identify string and prefix */
-	switch ( tag ) {
-	case DHCP_ISCSI_INITIATOR_IQN:
-		string = &iscsi_explicit_initiator_iqn;
-		break;
-	case DHCP_EB_USERNAME:
-		string = &iscsi_username;
-		break;
-	case DHCP_EB_PASSWORD:
-		string = &iscsi_password;
-		break;
-	case DHCP_HOST_NAME:
-		string = &iscsi_default_initiator_iqn;
-		prefix = "iqn.2000-09.org.etherboot:";
-		break;
-	default:
-		assert ( 0 );
-		return -EINVAL;
-	}
-
 	/* Free old string */
-	free ( *string );
-	*string = NULL;
+	free ( *setting->string );
+	*setting->string = NULL;
 
-	/* Allocate and fill new string */
-	prefix_len = strlen ( prefix );
-	len = ( prefix_len + option->len + 1 );
-	p = *string = malloc ( len );
+	/* Allocate new string */
+	prefix_len = strlen ( setting->prefix );
+	setting_len = fetch_setting_len ( NULL, setting->setting );
+	if ( setting_len < 0 ) {
+		/* Missing settings are not errors; leave strings as NULL */
+		return 0;
+	}
+	len = ( prefix_len + setting_len + 1 );
+	p = *setting->string = malloc ( len );
 	if ( ! p )
 		return -ENOMEM;
-	strcpy ( p, prefix );
-	dhcp_snprintf ( ( p + prefix_len ), ( len - prefix_len ), option );
+
+	/* Fill new string */
+	strcpy ( p, setting->prefix );
+	check_len = fetch_string_setting ( NULL, setting->setting,
+					   ( p + prefix_len ),
+					   ( len - prefix_len ) );
+	assert ( check_len == setting_len );
+
 	return 0;
 }
 
-/** DHCP iSCSI option applicators */
-struct dhcp_option_applicator dhcp_iscsi_applicators[] __dhcp_applicator = {
-	{
-		.tag = DHCP_ISCSI_INITIATOR_IQN,
-		.apply = apply_dhcp_iscsi_string,
-	},
-	{
-		.tag = DHCP_EB_USERNAME,
-		.apply = apply_dhcp_iscsi_string,
-	},
-	{
-		.tag = DHCP_EB_PASSWORD,
-		.apply = apply_dhcp_iscsi_string,
-	},
-	{
-		.tag = DHCP_HOST_NAME,
-		.apply = apply_dhcp_iscsi_string,
-	},
+/**
+ * Apply iSCSI settings
+ *
+ * @ret rc		Return status code
+ */
+static int apply_iscsi_settings ( void ) {
+	struct iscsi_string_setting *setting;
+	unsigned int i;
+	int rc;
+
+	for ( i = 0 ; i < ( sizeof ( iscsi_string_settings ) /
+			    sizeof ( iscsi_string_settings[0] ) ) ; i++ ) {
+		setting = &iscsi_string_settings[i];
+		if ( ( rc = apply_iscsi_string_setting ( setting ) ) != 0 ) {
+			DBG ( "iSCSI could not apply setting %s\n",
+			      setting->setting->name );
+			return rc;
+		}
+	}
+
+	return 0;
+}
+
+/** iSCSI settings applicator */
+struct settings_applicator iscsi_settings_applicator __settings_applicator = {
+	.apply = apply_iscsi_settings,
 };
 
 /****************************************************************************
